# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2017 Shai Seger <shaise at gmail>                       *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import Path
import Path.Base.Util as PathUtil
import Path.Dressup.Utils as PathDressup
import PathScripts.PathUtils as PathUtils
import Path.Main.Job as PathJob
import PathGui
import PathSimulator
import math
import os

from FreeCAD import Vector, Base

# lazily loaded modules
from lazy_loader.lazy_loader import LazyLoader

Mesh = LazyLoader("Mesh", globals(), "Mesh")
Part = LazyLoader("Part", globals(), "Part")

if FreeCAD.GuiUp:
    import FreeCADGui
    from PySide import QtGui, QtCore

_filePath = os.path.dirname(os.path.abspath(__file__))


class CAMSimTaskUi:
    def __init__(self, parent):
        # this will create a Qt widget from our ui file
        self.form = FreeCADGui.PySideUic.loadUi(":/panels/TaskPathSimulator.ui")
        self.parent = parent

    def accept(self):
        self.parent.accept()
        FreeCADGui.Control.closeDialog()

    def reject(self):
        self.parent.cancel()
        FreeCADGui.Control.closeDialog()


def TSError(msg):
    QtGui.QMessageBox.information(None, "Path Simulation", msg)


class PathSimulation:
    def __init__(self):
        self.debug = False
        self.timer = QtCore.QTimer()
        QtCore.QObject.connect(self.timer, QtCore.SIGNAL("timeout()"), self.PerformCut)
        self.stdrot = FreeCAD.Rotation(Vector(0, 0, 1), 0)
        self.iprogress = 0
        self.numCommands = 0
        self.simperiod = 20
        self.accuracy = 0.1
        self.resetSimulation = False
        self.jobs = []

    def Connect(self, but, sig):
        QtCore.QObject.connect(but, QtCore.SIGNAL("clicked()"), sig)

    def UpdateProgress(self):
        if self.numCommands > 0:
            self.taskForm.form.progressBar.setValue(self.iprogress * 100 / self.numCommands)

    def Activate(self):
        self.initdone = False
        self.taskForm = CAMSimTaskUi(self)
        form = self.taskForm.form
        self.Connect(form.toolButtonStop, self.SimStop)
        self.Connect(form.toolButtonPlay, self.SimPlay)
        self.Connect(form.toolButtonPause, self.SimPause)
        self.Connect(form.toolButtonStep, self.SimStep)
        self.Connect(form.toolButtonFF, self.SimFF)
        form.sliderSpeed.valueChanged.connect(self.onSpeedBarChange)
        self.onSpeedBarChange()
        form.sliderAccuracy.valueChanged.connect(self.onAccuracyBarChange)
        self.onAccuracyBarChange()
        self._populateJobSelection(form)
        form.comboJobs.currentIndexChanged.connect(self.onJobChange)
        self.onJobChange()
        FreeCADGui.Control.showDialog(self.taskForm)
        self.disableAnim = False
        self.isVoxel = True
        self.firstDrill = True
        self.voxSim = PathSimulator.PathSim()
        self.SimulateMill()
        self.initdone = True

    def _populateJobSelection(self, form):
        # Make Job selection combobox
        setJobIdx = 0
        jobName = ""
        jIdx = 0
        # Get list of Job objects in active document
        jobList = FreeCAD.ActiveDocument.findObjects("Path::FeaturePython", "Job.*")
        jCnt = len(jobList)

        # Check if user has selected a specific job for simulation
        guiSelection = FreeCADGui.Selection.getSelectionEx()
        if guiSelection:  #  Identify job selected by user
            sel = guiSelection[0]
            if hasattr(sel.Object, "Proxy") and isinstance(sel.Object.Proxy, PathJob.ObjectJob):
                jobName = sel.Object.Name
                FreeCADGui.Selection.clearSelection()

        # populate the job selection combobox
        form.comboJobs.blockSignals(True)
        form.comboJobs.clear()
        form.comboJobs.blockSignals(False)
        for j in jobList:
            form.comboJobs.addItem(j.ViewObject.Icon, j.Label)
            self.jobs.append(j)
            if j.Name == jobName or jCnt == 1:
                setJobIdx = jIdx
            jIdx += 1

        # Preselect GUI-selected job in the combobox
        if jobName or jCnt == 1:
            form.comboJobs.setCurrentIndex(setJobIdx)
        else:
            form.comboJobs.setCurrentIndex(0)

    def SetupSimulation(self):
        form = self.taskForm.form
        self.activeOps = []
        self.numCommands = 0
        self.ioperation = 0
        for i in range(form.listOperations.count()):
            if form.listOperations.item(i).checkState() == QtCore.Qt.CheckState.Checked:
                self.firstDrill = True
                self.activeOps.append(self.operations[i])
                self.numCommands += len(self.operations[i].Path.Commands)

        self.stock = self.job.Stock.Shape
        if self.isVoxel:
            maxlen = self.stock.BoundBox.XLength
            if maxlen < self.stock.BoundBox.YLength:
                maxlen = self.stock.BoundBox.YLength
            self.resolution = 0.01 * self.accuracy * maxlen
            self.voxSim.BeginSimulation(self.stock, self.resolution)
            (
                self.cutMaterial.Mesh,
                self.cutMaterialIn.Mesh,
            ) = self.voxSim.GetResultMesh()
        else:
            self.cutMaterial.Shape = self.stock
        self.busy = False
        self.tool = None
        for i in range(len(self.activeOps)):
            self.SetupOperation(0)
            if self.tool is not None:
                break
        self.iprogress = 0
        self.UpdateProgress()

    def SetupOperation(self, itool):
        self.operation = self.activeOps[itool]
        try:
            self.tool = PathDressup.toolController(self.operation).Tool
        except Exception:
            self.tool = None

        if self.tool is not None:
            self.cutTool.Shape = self.tool.Shape

            if not self.cutTool.Shape.isValid() or self.cutTool.Shape.isNull():
                self.EndSimulation()
                raise RuntimeError(
                    "Path Simulation: Error in tool geometry - {}".format(self.tool.Name)
                )

            self.cutTool.ViewObject.show()
            self.voxSim.SetToolShape(self.cutTool.Shape, 0.05 * self.accuracy)
        self.icmd = 0
        self.curpos = FreeCAD.Placement(self.initialPos, self.stdrot)
        self.cutTool.Placement = self.curpos
        self.opCommands = PathUtils.getPathWithPlacement(self.operation).Commands

    def SimulateMill(self):
        self.job = self.jobs[self.taskForm.form.comboJobs.currentIndex()]
        self.busy = False
        self.height = 10
        self.skipStep = False
        self.initialPos = Vector(0, 0, self.job.Stock.Shape.BoundBox.ZMax)
        # Add cut tool
        self.cutTool = FreeCAD.ActiveDocument.addObject("Part::FeaturePython", "CutTool")
        self.cutTool.ViewObject.Proxy = 0
        self.cutTool.ViewObject.hide()

        # Add cut material
        if self.isVoxel:
            self.cutMaterial = FreeCAD.ActiveDocument.addObject(
                "Mesh::FeaturePython", "CutMaterial"
            )
            self.cutMaterialIn = FreeCAD.ActiveDocument.addObject(
                "Mesh::FeaturePython", "CutMaterialIn"
            )
            self.cutMaterialIn.ViewObject.Proxy = 0
            self.cutMaterialIn.ViewObject.show()
            self.cutMaterialIn.ViewObject.ShapeColor = (1.0, 0.85, 0.45, 0.0)
        else:
            self.cutMaterial = FreeCAD.ActiveDocument.addObject(
                "Part::FeaturePython", "CutMaterial"
            )
            self.cutMaterial.Shape = self.job.Stock.Shape
        self.cutMaterial.ViewObject.Proxy = 0
        self.cutMaterial.ViewObject.show()
        self.cutMaterial.ViewObject.ShapeColor = (0.5, 0.25, 0.25, 0.0)

        # Add cut path solid for debug
        if self.debug:
            self.cutSolid = FreeCAD.ActiveDocument.addObject("Part::FeaturePython", "CutDebug")
            self.cutSolid.ViewObject.Proxy = 0
            self.cutSolid.ViewObject.hide()

        self.SetupSimulation()
        self.resetSimulation = True
        FreeCAD.ActiveDocument.recompute()

    def PerformCutBoolean(self):
        if self.resetSimulation:
            self.resetSimulation = False
            self.SetupSimulation()

        if self.busy:
            return
        self.busy = True

        cmd = self.opCommands[self.icmd]
        pathSolid = None

        if cmd.Name in ("G0", "G00"):
            self.firstDrill = True
            self.curpos = self.RapidMove(cmd, self.curpos)
        if cmd.Name in ("G1", "G01", "G2", "G02", "G3", "G03"):
            self.firstDrill = True
            if self.skipStep:
                self.curpos = self.RapidMove(cmd, self.curpos)
            else:
                (pathSolid, self.curpos) = self.GetPathSolid(self.tool, cmd, self.curpos)

        if cmd.Name == "G80":
            self.firstDrill = True
        if cmd.Name in ("G73", "G81", "G82", "G83"):
            if self.firstDrill:
                extendcommand = Path.Command("G0", {"Z": cmd.r})
                self.curpos = self.RapidMove(extendcommand, self.curpos)
                self.firstDrill = False
            extendcommand = Path.Command("G0", {"X": cmd.x, "Y": cmd.y, "Z": cmd.r})
            self.curpos = self.RapidMove(extendcommand, self.curpos)
            extendcommand = Path.Command("G1", {"X": cmd.x, "Y": cmd.y, "Z": cmd.z})
            self.curpos = self.RapidMove(extendcommand, self.curpos)
            extendcommand = Path.Command("G1", {"X": cmd.x, "Y": cmd.y, "Z": cmd.r})
            self.curpos = self.RapidMove(extendcommand, self.curpos)
        self.skipStep = False
        if pathSolid is not None:
            if self.debug:
                self.cutSolid.Shape = pathSolid
            newStock = self.stock.cut([pathSolid], 1e-3)
            try:
                if newStock.isValid():
                    self.stock = newStock.removeSplitter()
            except Exception:
                if self.debug:
                    print("invalid cut at cmd #{}".format(self.icmd))
        if not self.disableAnim:
            self.cutTool.Placement = FreeCAD.Placement(self.curpos, self.stdrot)
        self.icmd += 1
        self.iprogress += 1
        self.UpdateProgress()
        if self.icmd >= len(self.opCommands):
            self.ioperation += 1
            if self.ioperation >= len(self.activeOps):
                self.EndSimulation()
                return
            else:
                self.SetupOperation(self.ioperation)
        if not self.disableAnim:
            self.cutMaterial.Shape = self.stock
        self.busy = False

    def PerformCutVoxel(self):
        if self.resetSimulation:
            self.resetSimulation = False
            self.SetupSimulation()

        if self.busy:
            return
        self.busy = True

        cmd = self.opCommands[self.icmd]
        # for cmd in job.Path.Commands:
        if cmd.Name in ("G0", "G00", "G1", "G01", "G2", "G02", "G3", "G03"):
            index = self.icmd
            if cmd.Name in ("G2", "G02", "G3", "G03"):
                while cmd.z is None:
                    index -= 1
                    if index < 0:
                        cmd.z = 0
                        break
                    cmd.z = self.opCommands[index].z

            self.firstDrill = True
            if cmd.Name in ("G2", "G02", "G3", "G03") and (cmd.k or 0) == 0:
                cx = self.curpos.Base.x + (cmd.i or 0)
                cy = self.curpos.Base.y + (cmd.j or 0)
                a0 = math.atan2(self.curpos.Base.y - cy, self.curpos.Base.x - cx)
                a1 = math.atan2(cmd.y - cy, cmd.x - cx)
                da = a1 - a0
                if cmd.Name in ("G3", "G03"):
                    da = da % (2 * math.pi)
                else:
                    da = -((-da) % (2 * math.pi))
                r = math.sqrt((cmd.i or 0) ** 2 + (cmd.j or 0) ** 2)
                n = math.ceil(math.sqrt(r / self.resolution * da * da))
                da = da / n
                dz = (cmd.z - self.curpos.Base.z) / n
                cmd.Name = "G1"
                for i in range(n):
                    a0 += da
                    cmd.x = cx + r * math.cos(a0)
                    cmd.y = cy + r * math.sin(a0)
                    cmd.z = self.curpos.Base.z + dz
                    self.curpos = self.voxSim.ApplyCommand(self.curpos, cmd)
            else:
                self.curpos = self.voxSim.ApplyCommand(self.curpos, cmd)
            if not self.disableAnim:
                self.cutTool.Placement = self.curpos
                (
                    self.cutMaterial.Mesh,
                    self.cutMaterialIn.Mesh,
                ) = self.voxSim.GetResultMesh()
        if cmd.Name == "G80":
            self.firstDrill = True
        if cmd.Name in ("G73", "G81", "G82", "G83"):
            extendcommands = []
            if self.firstDrill:
                extendcommands.append(Path.Command("G0", {"Z": cmd.r}))
                self.firstDrill = False
            extendcommands.append(Path.Command("G0", {"X": cmd.x, "Y": cmd.y, "Z": cmd.r}))
            extendcommands.append(Path.Command("G1", {"X": cmd.x, "Y": cmd.y, "Z": cmd.z}))
            extendcommands.append(Path.Command("G1", {"X": cmd.x, "Y": cmd.y, "Z": cmd.r}))
            for ecmd in extendcommands:
                self.curpos = self.voxSim.ApplyCommand(self.curpos, ecmd)
                if not self.disableAnim:
                    self.cutTool.Placement = self.curpos
                    (
                        self.cutMaterial.Mesh,
                        self.cutMaterialIn.Mesh,
                    ) = self.voxSim.GetResultMesh()
        self.icmd += 1
        self.iprogress += 1
        self.UpdateProgress()
        if self.icmd >= len(self.opCommands):
            self.ioperation += 1
            if self.ioperation >= len(self.activeOps):
                self.EndSimulation()
                return
            else:
                self.SetupOperation(self.ioperation)
        self.busy = False

    def PerformCut(self):
        if self.isVoxel:
            self.PerformCutVoxel()
        else:
            self.PerformCutBoolean()

    def RapidMove(self, cmd, curpos):
        path = Path.Geom.edgeForCmd(cmd, curpos)  # hack to overcome occ bug
        if path is None:
            return curpos
        return path.valueAt(path.LastParameter)

    # get a solid representation of a tool going along path
    def GetPathSolid(self, tool, cmd, pos):
        toolPath = Path.Geom.edgeForCmd(cmd, pos)
        startDir = toolPath.tangentAt(0)
        startDir[2] = 0.0
        endPos = toolPath.valueAt(toolPath.LastParameter)
        endDir = toolPath.tangentAt(toolPath.LastParameter)
        try:
            startDir.normalize()
            endDir.normalize()
        except Exception:
            return (None, endPos)

        # hack to overcome occ bugs
        rad = float(tool.Diameter) / 2.0 - 0.001 * pos[2]
        if type(toolPath.Curve) is Part.Circle and toolPath.Curve.Radius <= rad:
            rad = toolPath.Curve.Radius - 0.01 * (pos[2] + 1)
            return (None, endPos)

        # create the path shell
        toolProf = self.CreateToolProfile(tool, startDir, pos, rad)
        rotmat = Base.Matrix()
        rotmat.move(pos.negative())
        rotmat.rotateZ(math.pi)
        rotmat.move(pos)
        mirroredProf = toolProf.transformGeometry(rotmat)
        fullProf = Part.Wire([toolProf, mirroredProf])
        pathWire = Part.Wire(toolPath)
        try:
            pathShell = pathWire.makePipeShell([fullProf], False, True)
        except Exception:
            if self.debug:
                Part.show(pathWire)
                Part.show(fullProf)
            return (None, endPos)

        # create the start cup
        startCup = toolProf.revolve(pos, Vector(0, 0, 1), -180)

        # create the end cup
        endProf = self.CreateToolProfile(tool, endDir, endPos, rad)
        endCup = endProf.revolve(endPos, Vector(0, 0, 1), 180)

        fullShell = Part.makeShell(startCup.Faces + pathShell.Faces + endCup.Faces)
        return (Part.makeSolid(fullShell).removeSplitter(), endPos)

    # create radial profile of the tool (90 degrees to the direction of the path)
    def CreateToolProfile(self, tool, dir, pos, rad):
        type = tool.ToolType
        xf = dir[0] * rad
        yf = dir[1] * rad
        xp = pos[0]
        yp = pos[1]
        zp = pos[2]
        h = tool.CuttingEdgeHeight
        if h <= 0.0:  # set default if user fails to avoid freeze
            h = 1.0
            Path.Log.error("SET Tool Length")
        # common to all tools
        vTR = Vector(xp + yf, yp - xf, zp + h)
        vTC = Vector(xp, yp, zp + h)
        vBC = Vector(xp, yp, zp)
        lT = Part.makeLine(vTR, vTC)
        res = None
        if type == "ChamferMill":
            ang = 90 - tool.CuttingEdgeAngle / 2.0
            if ang > 80:
                ang = 80
            if ang < 0:
                ang = 0
            h1 = math.tan(ang * math.pi / 180) * rad
            if h1 > (h - 0.1):
                h1 = h - 0.1
            vBR = Vector(xp + yf, yp - xf, zp + h1)
            lR = Part.makeLine(vBR, vTR)
            lB = Part.makeLine(vBC, vBR)
            res = Part.Wire([lB, lR, lT])

        elif type == "BallEndMill":
            h1 = rad
            if h1 >= h:
                h1 = h - 0.1
            vBR = Vector(xp + yf, yp - xf, zp + h1)
            r2 = h1 / 2.0
            h2 = rad - math.sqrt(rad * rad - r2 * r2)
            vBCR = Vector(xp + yf / 2.0, yp - xf / 2.0, zp + h2)
            cB = Part.Edge(Part.Arc(vBC, vBCR, vBR))
            lR = Part.makeLine(vBR, vTR)
            res = Part.Wire([cB, lR, lT])

        else:  # default: assume type == "EndMill"
            vBR = Vector(xp + yf, yp - xf, zp)
            lR = Part.makeLine(vBR, vTR)
            lB = Part.makeLine(vBC, vBR)
            res = Part.Wire([lB, lR, lT])

        return res

    def onJobChange(self):
        form = self.taskForm.form
        j = self.jobs[form.comboJobs.currentIndex()]
        self.job = j
        form.listOperations.clear()
        self.operations = []
        for op in j.Operations.OutList:
            if PathUtil.opProperty(op, "Active"):
                listItem = QtGui.QListWidgetItem(op.ViewObject.Icon, op.Label)
                listItem.setFlags(listItem.flags() | QtCore.Qt.ItemIsUserCheckable)
                listItem.setCheckState(QtCore.Qt.CheckState.Checked)
                self.operations.append(op)
                form.listOperations.addItem(listItem)
        if self.initdone:
            self.SetupSimulation()

    def onSpeedBarChange(self):
        form = self.taskForm.form
        self.simperiod = 1000 / form.sliderSpeed.value()
        form.labelGPerSec.setText(str(form.sliderSpeed.value()) + " G/s")
        self.timer.setInterval(self.simperiod)

    def onAccuracyBarChange(self):
        form = self.taskForm.form
        self.accuracy = 1.1 - 0.1 * form.sliderAccuracy.value()
        form.labelAccuracy.setText(str(round(self.accuracy, 1)) + "%")

    def GuiBusy(self, isBusy):
        form = self.taskForm.form
        form.toolButtonPlay.setEnabled(not isBusy)
        form.toolButtonPause.setEnabled(isBusy)
        form.toolButtonStep.setEnabled(not isBusy)
        form.toolButtonFF.setEnabled(not isBusy)

    def EndSimulation(self):
        self.UpdateProgress()
        self.timer.stop()
        self.GuiBusy(False)
        self.ViewShape()
        self.resetSimulation = True

    def SimStop(self):
        self.cutTool.ViewObject.hide()
        self.iprogress = 0
        self.EndSimulation()

    def InvalidOperation(self):
        if len(self.activeOps) == 0:
            return True
        if self.tool is None:
            TSError("No tool assigned for the operation")
            return True
        return False

    def SimFF(self):
        if self.InvalidOperation():
            return
        self.GuiBusy(True)
        self.timer.start(1)
        self.disableAnim = True

    def SimStep(self):
        if self.InvalidOperation():
            return
        self.disableAnim = False
        self.PerformCut()

    def SimPlay(self):
        if self.InvalidOperation():
            return
        self.disableAnim = False
        self.GuiBusy(True)
        self.timer.start(self.simperiod)

    def ViewShape(self):
        if self.isVoxel:
            (
                self.cutMaterial.Mesh,
                self.cutMaterialIn.Mesh,
            ) = self.voxSim.GetResultMesh()
        else:
            self.cutMaterial.Shape = self.stock

    def SimPause(self):
        if self.disableAnim:
            self.ViewShape()
        self.GuiBusy(False)
        self.timer.stop()

    def RemoveTool(self):
        if self.cutTool is None:
            return
        FreeCAD.ActiveDocument.removeObject(self.cutTool.Name)
        self.cutTool = None

    def RemoveInnerMaterial(self):
        if self.cutMaterialIn is not None:
            if self.isVoxel and self.cutMaterial is not None:
                mesh = Mesh.Mesh()
                mesh.addMesh(self.cutMaterial.Mesh)
                mesh.addMesh(self.cutMaterialIn.Mesh)
                self.cutMaterial.Mesh = mesh
            FreeCAD.ActiveDocument.removeObject(self.cutMaterialIn.Name)
            self.cutMaterialIn = None

    def RemoveMaterial(self):
        if self.cutMaterial is not None:
            FreeCAD.ActiveDocument.removeObject(self.cutMaterial.Name)
            self.cutMaterial = None
        self.RemoveInnerMaterial()

    def accept(self):
        self.EndSimulation()
        self.RemoveInnerMaterial()
        self.RemoveTool()

    def cancel(self):
        self.EndSimulation()
        self.RemoveTool()
        self.RemoveMaterial()


class CommandPathSimulate:
    def GetResources(self):
        return {
            "Pixmap": "CAM_Simulator",
            "MenuText": QtCore.QT_TRANSLATE_NOOP("CAM_Simulator", "Legacy CAM Simulator"),
            "Accel": "P, M",
            "ToolTip": QtCore.QT_TRANSLATE_NOOP("CAM_Simulator", "Simulates G-code on stock"),
        }

    def IsActive(self):
        if FreeCAD.ActiveDocument is not None:
            for o in FreeCAD.ActiveDocument.Objects:
                if o.Name[:3] == "Job":
                    return True
        return False

    def Activated(self):
        pathSimulation = PathSimulation()
        pathSimulation.Activate()


if FreeCAD.GuiUp:
    # register the FreeCAD command
    FreeCADGui.addCommand("CAM_Simulator", CommandPathSimulate())
    FreeCAD.Console.PrintLog("Loading PathSimulator Gui… done\n")



from Base.Metadata import export
from Base.Vector import Vector
from Conic import Conic
from Point import Point
from typing import overload

@export(
    PythonName="Part.Circle",
    Twin="GeomCircle",
    TwinPointer="GeomCircle",
    Include="Mod/Part/App/Geometry.h",
    FatherInclude="Mod/Part/App/ConicPy.h",
    FatherNamespace="Part",
    Constructor=True,
)
class Circle(Conic):
    """
    Describes a circle in 3D space
    To create a circle there are several ways:
    Part.Circle()
        Creates a default circle with center (0,0,0) and radius 1

    Part.Circle(Circle)
        Creates a copy of the given circle

    Part.Circle(Circle, Distance)
        Creates a circle parallel to given circle at a certain distance

    Part.Circle(Center,Normal,Radius)
        Creates a circle defined by center, normal direction and radius

    Part.Circle(Point1,Point2,Point3)
        Creates a circle defined by three non-linear points
    """

    Radius: float = 0.0
    """The radius of the circle."""

    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, circle: "Circle") -> None: ...
    @overload
    def __init__(self, circle: "Circle", distance: float) -> None: ...
    @overload
    def __init__(self, center: Point, normal: Vector, radius: float) -> None: ...
    @overload
    def __init__(self, point1: Point, point2: Point, point3: Point) -> None: ...



from typing import Any, Final

from Base.PyObjectBase import PyObjectBase
from Base.Metadata import constmethod, export

@export(
    Include="Mod/TechDraw/App/Cosmetic.h",
    Namespace="TechDraw",
    Constructor=True,
    Delete=True,
)
class GeomFormat(PyObjectBase):
    """
    GeomFormat specifies appearance parameters for TechDraw Geometry objects
    
    Author: WandererFan (wandererfan@gmail.com)
    License: LGPL-2.1-or-later
    """

    @constmethod
    def clone(self) -> Any:
        """Create a clone of this geomformat"""
        ...

    @constmethod
    def copy(self) -> Any:
        """Create a copy of this geomformat"""
        ...
    Tag: Final[str]
    """Gives the tag of the GeomFormat as string."""



from Base.Metadata import export, constmethod, class_declarations
from Base.BaseClass import BaseClass
from typing import Any, Final, List


@export(
    Include="App/StringHasher.h",
    Reference=True,
)
@class_declarations("""private:
    friend class StringID;
    int _index = 0;
        """
)
class StringID(BaseClass):
    """
    This is the StringID class

    Author: Zheng, Lei (realthunder.dev@gmail.com)
    Licence: LGPL
    """

    @constmethod
    def isSame(self, other: "StringID") -> bool:
        """
        Check if two StringIDs are the same
        """
        ...

    Value: Final[int] = 0
    """Return the integer value of this ID"""

    Related: Final[List[Any]] = []
    """Return the related string IDs"""

    Data: Final[str] = ""
    """Return the data associated with this ID"""

    IsBinary: Final[bool] = False
    """Check if the data is binary,"""

    IsHashed: Final[bool] = False
    """Check if the data is hash, if so 'Data' returns a base64 encoded string of the raw hash"""

    Index: int = 0
    """Geometry index. Only meaningful for geometry element name"""



# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileCopyrightText: 2022 FreeCAD Project Association
# SPDX-FileNotice: Part of the AddonManager.

# ***************************************************************************
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""Defines a QWidget-derived class for displaying the cache load status."""

import os

import addonmanager_freecad_interface as fci

translate = fci.translate

# Get whatever version of PySide we can
try:
    from PySide import QtCore, QtGui, QtWidgets  # Use the FreeCAD wrapper
except ImportError:
    try:
        from PySide6 import QtCore, QtGui, QtWidgets  # Outside FreeCAD, try Qt6 first
    except ImportError:
        from PySide2 import QtCore, QtGui, QtWidgets  # Fall back to Qt5

_TOTAL_INCREMENTS = 1000


class Progress:
    """Represents progress through a process composed of multiple sub-tasks."""

    def __init__(
        self,
        *,
        status_text: str = "",
        number_of_tasks: int = 1,
        current_task: int = 0,
        current_task_progress: float = 0.0,
    ):
        if number_of_tasks < 1:
            raise ValueError(f"Number of tasks must be at least one, not {number_of_tasks}")
        if current_task < 0 or current_task >= number_of_tasks:
            raise ValueError(
                "Current task must be between 0 and the number of tasks "
                f"({number_of_tasks}), not {current_task}"
            )
        if current_task_progress < 0.0:
            current_task_progress = 0.0
        elif current_task_progress > 100.0:
            current_task_progress = 100.0
        self.status_text: str = status_text
        self._number_of_tasks: int = number_of_tasks
        self._current_task: int = current_task
        self._current_task_progress: float = current_task_progress

    @property
    def number_of_tasks(self):
        return self._number_of_tasks

    @number_of_tasks.setter
    def number_of_tasks(self, value: int):
        if not isinstance(value, int):
            raise TypeError("Number of tasks must be an integer")
        if value < 1:
            raise ValueError("Number of tasks must be at least one")
        self._number_of_tasks = value

    @property
    def current_task(self):
        """The current task (zero-indexed, always less than the number of tasks)"""
        return self._current_task

    @current_task.setter
    def current_task(self, value: int):
        if not isinstance(value, int):
            raise TypeError("Current task must be an integer")
        if value < 0:
            raise ValueError("Current task must be at least zero")
        if value >= self._number_of_tasks:
            raise ValueError("Current task must be less than the total number of tasks")
        self._current_task = value

    @property
    def current_task_progress(self):
        """Current task progress, guaranteed to be in the range [0.0, 100.0]. Attempts to set a
        value outside that range are clamped to the range."""
        return self._current_task_progress

    @current_task_progress.setter
    def current_task_progress(self, value: float):
        """Set the current task's progress. Rather than raising an exception when the value is
        outside the expected range of [0,100], clamp the task progress to allow for some
        floating point imprecision in its calculation."""
        if value < 0.0:
            value = 0.0
        elif value > 100.0:
            value = 100.0
        self._current_task_progress = value

    def next_task(self) -> None:
        """Increment the task counter and reset the progress"""
        self.current_task += 1
        self.current_task_progress = 0.0

    def overall_progress(self) -> float:
        """Gets the overall progress as a fractional value in the range [0, 1]"""
        base = self._current_task / self._number_of_tasks
        fraction = self._current_task_progress / (100.0 * self._number_of_tasks)
        return base + fraction


class WidgetProgressBar(QtWidgets.QWidget):
    """A multipart progress bar widget, including a stop button and a status label."""

    stop_clicked = QtCore.Signal()

    def __init__(self, parent: QtWidgets.QWidget = None):
        super().__init__(parent)
        self.vertical_layout = None
        self.horizontal_layout = None
        self.progress_bar = None
        self.status_label = None
        self.stop_button = None
        self._setup_ui()

    def _setup_ui(self):
        self.vertical_layout = QtWidgets.QVBoxLayout(self)
        self.horizontal_layout = QtWidgets.QHBoxLayout()
        self.progress_bar = QtWidgets.QProgressBar(self)
        self.status_label = QtWidgets.QLabel(self)
        self.stop_button = QtWidgets.QToolButton(self)
        self.progress_bar.setMaximum(_TOTAL_INCREMENTS)
        self.stop_button.clicked.connect(self.stop_clicked)
        icon_path = os.path.join(os.path.dirname(__file__), "..", "Resources", "icons")
        self.stop_button.setIcon(
            QtGui.QIcon.fromTheme("stop", QtGui.QIcon(os.path.join(icon_path, "debug-stop.svg")))
        )
        self.vertical_layout.addLayout(self.horizontal_layout)
        self.vertical_layout.addWidget(self.status_label)
        self.horizontal_layout.addWidget(self.progress_bar)
        self.horizontal_layout.addWidget(self.stop_button)
        self.vertical_layout.setContentsMargins(0, 0, 0, 0)
        self.horizontal_layout.setContentsMargins(0, 0, 0, 0)

    def set_progress(self, progress: Progress) -> None:
        self.status_label.setText(progress.status_text)
        self.progress_bar.setValue(progress.overall_progress() * _TOTAL_INCREMENTS)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2017 Lorenz Lechner                                     *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This library is free software; you can redistribute it and/or         *
# *   modify it under the terms of the GNU Library General Public           *
# *   License as published by the Free Software Foundation; either          *
# *   version 2 of the License, or (at your option) any later version.      *
# *                                                                         *
# *   This library  is distributed in the hope that it will be useful,      *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this library; see the file COPYING.LIB. If not,    *
# *   write to the Free Software Foundation, Inc., 59 Temple Place,         *
# *   Suite 330, Boston, MA  02111-1307, USA                                *
# *                                                                         *
# ***************************************************************************/

import Mesh
import FreeCAD as App
import FreeCADGui as Gui
import Part
import MeshPartGui

from PySide.QtCore import QT_TRANSLATE_NOOP  # for translations


class BaseCommand(object):
    def __init__(self):
        pass

    def IsActive(self):
        if App.ActiveDocument is None:
            return False
        else:
            return True


class CreateFlatMesh(BaseCommand):
    """create flat wires from a meshed face"""

    def GetResources(self):
        return {
            "Pixmap": "MeshPart_CreateFlatMesh.svg",
            "MenuText": QT_TRANSLATE_NOOP("MeshPart_CreateFlatMesh", "Unwrap Mesh"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "MeshPart_CreateFlatMesh", "Finds a flat representation of a mesh"
            ),
        }

    def Activated(self):
        import numpy as np
        import flatmesh

        obj = Gui.Selection.getSelection()[
            0
        ]  # obj must be a Mesh (Mesh-Design->Meshes->Create-Mesh)
        points = np.array([[i.x, i.y, i.z] for i in obj.Mesh.Points])
        faces = np.array([list(i) for i in obj.Mesh.Topology[1]])
        flattener = flatmesh.FaceUnwrapper(points, faces)
        flattener.findFlatNodes(5, 0.95)
        boundaries = flattener.getFlatBoundaryNodes()
        # print('number of nodes: {}'.format(len(flattener.ze_nodes)))
        # print('number of faces: {}'.format(len(flattener.tris)))

        wires = []
        for edge in boundaries:
            pi = Part.makePolygon([App.Vector(*node) for node in edge])
            Part.show(Part.Wire(pi))

    def IsActive(self):
        assert super(CreateFlatMesh, self).IsActive()
        assert isinstance(Gui.Selection.getSelection()[0].Mesh, Mesh.Mesh)
        return True


class CreateFlatFace(BaseCommand):
    """create a flat face from a single face
    only full faces are supported right now"""

    def GetResources(self):
        return {
            "Pixmap": "MeshPart_CreateFlatFace.svg",
            "MenuText": QT_TRANSLATE_NOOP("MeshPart_CreateFlatFace", "Unwrap Face"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "MeshPart_CreateFlatFace", "Finds a flat representation of a face"
            ),
        }

    def Activated(self):
        import numpy as np
        import flatmesh

        face = Gui.Selection.getSelectionEx()[0].SubObjects[0]
        shape = face.toNurbs()
        face = shape.Faces[0]
        nurbs = face.Surface
        nurbs.setUNotPeriodic()
        nurbs.setVNotPeriodic()
        bs = nurbs.toBSpline(1, "C0", "C0", 3, 3, 10)
        face = bs.toShape()
        face.tessellate(0.01)
        flattener = flatmesh.FaceUnwrapper(face)
        flattener.findFlatNodes(5, 0.99)
        poles = flattener.interpolateFlatFace(face)
        num_u_poles = len(bs.getPoles())
        num_v_poles = len(bs.getPoles()[0])
        i = 0
        for u in range(num_u_poles):
            for v in range(num_v_poles):
                bs.setPole(u + 1, v + 1, App.Vector(poles[i]))
                i += 1
        Part.show(bs.toShape())

    def IsActive(self):
        assert super(CreateFlatFace, self).IsActive()
        assert isinstance(Gui.Selection.getSelectionEx()[0].SubObjects[0], Part.Face)
        return True


# Test if pybind11 dependency is available
try:
    import flatmesh

    Gui.addCommand("MeshPart_CreateFlatMesh", CreateFlatMesh())
    Gui.addCommand("MeshPart_CreateFlatFace", CreateFlatFace())
except ImportError:
    App.Console.PrintLog("flatmesh-commands are not available\n")
    App.Console.PrintLog("flatmesh needs pybind11 as build dependency\n")
except AttributeError:
    # Can happen when running FreeCAD in headless mode
    pass



from Base.Metadata import export, constmethod
from Part.App.Geom2d.Conic2d import Conic2d
from typing import Final

@export(
    Include="Mod/Part/App/Geometry2d.h",
    FatherInclude="Mod/Part/App/Geom2d/Conic2dPy.h",
    Twin="Geom2dParabola",
    TwinPointer="Geom2dParabola",
    PythonName="Part.Geom2d.Parabola2d",
    Constructor=True,
)
class Parabola2d(Conic2d):
    """
    Describes a parabola in 2D space

    Author: Werner Mayer (wmayer@users.sourceforge.net)
    Licence: LGPL
    """

    Focal: float = ...
    """
    The focal distance is the distance between the apex and the focus of the parabola.
    """

    Focus: Final[object] = ...
    """
    The focus is on the positive side of the
    'X Axis' of the local coordinate system of the parabola.
    """

    Parameter: Final[float] = ...
    """
    Compute the parameter of this parabola which is the distance between its focus
    and its directrix. This distance is twice the focal length.
    """



from Base.Metadata import export, constmethod
from Base.Persistence import Persistence
from App.Extension import Extension
from Base.Vector import Vector
from Base.Matrix import Matrix
from typing import Final, List, Optional

@export(
    Include="Mod/Part/App/Geometry.h",
    Constructor=True,
    Delete=True,
)
class Geometry(Persistence):
    """
    The abstract class Geometry for 3D space is the root class of all geometric objects.
    It describes the common behavior of these objects when:
    - applying geometric transformations to objects, and
    - constructing objects by geometric transformation (including copying).
    """

    Tag: Final[str]
    """Gives the tag of the geometry as string."""

    def mirror(self, geometry: "Geometry") -> None:
        """
        Performs the symmetrical transformation of this geometric object
        """
        ...

    def rotate(self, angle: float, axis: Vector) -> None:
        """
        Rotates this geometric object at angle Ang (in radians) about axis
        """
        ...

    def scale(self, center: Vector, factor: float) -> None:
        """
        Applies a scaling transformation on this geometric object with a center and scaling factor
        """
        ...

    def transform(self, transformation: Matrix) -> None:
        """
        Applies a transformation to this geometric object
        """
        ...

    def translate(self, vector: Vector) -> None:
        """
        Translates this geometric object
        """
        ...

    @constmethod
    def copy(self) -> "Geometry":
        """
        Create a copy of this geometry
        """
        ...

    @constmethod
    def clone(self) -> "Geometry":
        """
        Create a clone of this geometry with the same Tag
        """
        ...

    @constmethod
    def isSame(self, geom: "Geometry", tol: float, angulartol: float) -> bool:
        """
        isSame(geom, tol, angulartol) -> boolean

        Compare this geometry to another one
        """
        ...

    @constmethod
    def hasExtensionOfType(self, type_name: str) -> bool:
        """
        Returns a boolean indicating whether a geometry extension of the type indicated as a string exists.
        """
        ...

    @constmethod
    def hasExtensionOfName(self, name: str) -> bool:
        """
        Returns a boolean indicating whether a geometry extension with the name indicated as a string exists.
        """
        ...

    @constmethod
    def getExtensionOfType(self, type_name: str) -> Optional[Extension]:
        """
        Gets the first geometry extension of the type indicated by the string.
        """
        ...

    @constmethod
    def getExtensionOfName(self, name: str) -> Optional[Extension]:
        """
        Gets the first geometry extension of the name indicated by the string.
        """
        ...

    def setExtension(self, extension: Extension) -> None:
        """
        Sets a geometry extension of the indicated type.
        """
        ...

    def deleteExtensionOfType(self, type_name: str) -> None:
        """
        Deletes all extensions of the indicated type.
        """
        ...

    def deleteExtensionOfName(self, name: str) -> None:
        """
        Deletes all extensions of the indicated name.
        """
        ...

    @constmethod
    def getExtensions(self) -> List[Extension]:
        """
        Returns a list with information about the geometry extensions.
        """
        ...



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2025 Furgo                                              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

import Arch
import Draft
import os
import FreeCAD as App
from bimtests import TestArchBase


class TestArchSectionPlane(TestArchBase.TestArchBase):

    def test_makeSectionPlane(self):
        """Test the makeSectionPlane function."""
        operation = "Testing makeSectionPlane function"
        self.printTestMessage(operation)

        section_plane = Arch.makeSectionPlane(name="TestSectionPlane")
        self.assertIsNotNone(
            section_plane, "makeSectionPlane failed to create a section plane object."
        )
        self.assertEqual(
            section_plane.Label, "TestSectionPlane", "Section plane label is incorrect."
        )

    def testViewGeneration(self):
        """Tests the whole TD view generation workflow"""

        # Create a few objects
        points = [App.Vector(0.0, 0.0, 0.0), App.Vector(2000.0, 0.0, 0.0)]
        line = Draft.make_wire(points)
        wall = Arch.makeWall(line, height=2000)
        wpl = App.Placement(App.Vector(500, 0, 1500), App.Vector(1, 0, 0), -90)
        win = Arch.makeWindowPreset(
            "Fixed",
            width=1000.0,
            height=1000.0,
            h1=50.0,
            h2=50.0,
            h3=50.0,
            w1=100.0,
            w2=50.0,
            o1=0.0,
            o2=50.0,
            placement=wpl,
        )
        win.Hosts = [wall]
        profile = Arch.makeProfile([169, "HEA", "HEA100", "H", 100.0, 96.0, 5.0, 8.0])
        column = Arch.makeStructure(profile, height=2000.0)
        column.Profile = "HEA100"
        column.Placement.Base = App.Vector(500.0, 600.0, 0.0)
        level = Arch.makeFloor()
        level.addObjects([wall, column])
        App.ActiveDocument.recompute()

        # Create a drawing view
        section = Arch.makeSectionPlane(level)
        drawing = Arch.make2DDrawing()
        view = Draft.make_shape2dview(section)
        cut = Draft.make_shape2dview(section)
        cut.InPlace = False
        cut.ProjectionMode = "Cutfaces"
        drawing.addObjects([view, cut])
        App.ActiveDocument.recompute()

        # Create a TD page
        tpath = os.path.join(
            App.getResourceDir(), "Mod", "TechDraw", "Templates", "ISO", "A3_Landscape_blank.svg"
        )
        page = App.ActiveDocument.addObject("TechDraw::DrawPage", "Page")
        template = App.ActiveDocument.addObject("TechDraw::DrawSVGTemplate", "Template")
        template.Template = tpath
        page.Template = template
        view = App.ActiveDocument.addObject("TechDraw::DrawViewDraft", "DraftView")
        view.Source = drawing
        page.addView(view)
        view.Scale = 1.0
        view.X = "20cm"
        view.Y = "15cm"
        App.ActiveDocument.recompute()
        assert True



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   (c) 2019 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de>           *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides functions to create polar Array objects."""
## @package make_polararray
# \ingroup draftmake
# \brief Provides functions to create polar Array objects.

## \addtogroup draftmake
# @{
import FreeCAD as App
import draftutils.utils as utils
import draftmake.make_array as make_array

from draftutils.messages import _err
from draftutils.translate import translate


def make_polar_array(base_object, number=5, angle=360, center=App.Vector(0, 0, 0), use_link=True):
    """Create a polar array from the given object.

    Parameters
    ----------
    base_object: Part::Feature or str
        Any of object that has a `Part::TopoShape` that can be duplicated.
        This means most 2D and 3D objects produced with any workbench.
        If it is a string, it must be the `Label` of that object.
        Since a label is not guaranteed to be unique in a document,
        it will use the first object found with this label.

    number: int, optional
        It defaults to 5.
        The number of copies produced in the polar pattern.

    angle: float, optional
        It defaults to 360.
        The magnitude in degrees swept by the polar pattern.

    center: Base::Vector3, optional
        It defaults to the origin `App.Vector(0, 0, 0)`.
        The vector indicating the center of rotation of the array.

    use_link: bool, optional
        It defaults to `True`.
        If it is `True` the produced copies are not `Part::TopoShape` copies,
        but rather `App::Link` objects.
        The Links repeat the shape of the original `obj` exactly,
        and therefore the resulting array is more memory efficient.

        Also, when `use_link` is `True`, the `Fuse` property
        of the resulting array does not work; the array doesn't
        contain separate shapes, it only has the original shape repeated
        many times, so there is nothing to fuse together.

        If `use_link` is `False` the original shape is copied many times.
        In this case the `Fuse` property is able to fuse
        all copies into a single object, if they touch each other.

    Returns
    -------
    Part::FeaturePython
        A scripted object of type `'Array'`.
        Its `Shape` is a compound of the copies of the original object.

    None
        If there is a problem it will return `None`.

    See Also
    --------
    make_ortho_array, make_circular_array, make_path_array, make_point_array
    """
    _name = "make_polar_array"

    found, base_object = utils.find_object(base_object, doc=App.activeDocument())
    if not found:
        _err(translate("draft", "Wrong input: base_object not in document."))
        return None

    try:
        utils.type_check([(number, int)], name=_name)
    except TypeError:
        _err(translate("draft", "Wrong input: must be an integer number."))
        return None

    try:
        utils.type_check([(angle, (int, float))], name=_name)
    except TypeError:
        _err(translate("draft", "Wrong input: must be a number."))
        return None

    try:
        utils.type_check([(center, App.Vector)], name=_name)
    except TypeError:
        _err(translate("draft", "Wrong input: must be a vector."))
        return None

    use_link = bool(use_link)
    new_obj = make_array.make_array(
        base_object, arg1=center, arg2=angle, arg3=number, use_link=use_link
    )
    return new_obj


## @}



from Base.Metadata import export, constmethod
from Base.BaseClass import BaseClass
from typing import Any, Final, overload, Dict


@export(
    Constructor=True,
    Reference=True,
)
class StringHasher(BaseClass):
    """
    This is the StringHasher class

    Author: Zheng, Lei (realthunder.dev@gmail.com)
    Licence: LGPL
    """

    @overload
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        ...

    def getID(self, arg: Any, base64: bool = False) -> Any:
        """
        getID(txt|id, base64=False) -> StringID

        If the input is text, return a StringID object that is unique within this hasher. This
        StringID object is reference counted. The hasher may only save hash ID's that are used.

        If the input is an integer, then the hasher will try to find the StringID object stored
        with the same integer value.

        base64: indicate if the input 'txt' is base64 encoded binary data
        """
        ...

    @overload
    def getID(self, txt: str, base64: bool = False) -> Any: ...

    @overload
    def getID(self, id: int, base64: bool = False) -> Any: ...

    @constmethod
    def isSame(self, other: "StringHasher") -> bool:
        """
        Check if two hasher are the same
        """
        ...

    Count: Final[int] = 0
    """Return count of used hashes"""

    Size: Final[int] = 0
    """Return the size of the hashes"""

    SaveAll: bool = False
    """Whether to save all string hashes regardless of its use count"""

    Threshold: int = 0
    """Data length exceed this threshold will be hashed before storing"""

    Table: Final[Dict[int, str]] = {}
    """Return the entire string table as Int->String dictionary"""



# ***************************************************************************
# *   Copyright (c) 2021 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "Mystran add femelement geometry"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

## \addtogroup FEM
#  @{


def add_femelement_geometry(f, model, mystran_writer):

    # generate pyNastran code
    # HACK, the if statemant needs improvement, see calculix solver
    if mystran_writer.member.geos_beamsection:
        beamsec_obj = mystran_writer.member.geos_beamsection[0]["Object"]
        if beamsec_obj.SectionType == "Rectangular":
            height = beamsec_obj.RectHeight.getValueAs("mm").Value
            width = beamsec_obj.RectWidth.getValueAs("mm").Value
            pynas_code = "# pbarl card, properties of a simple beam element (CBAR entry)\n"
            pynas_code += "# defined by cross-sectional dimensions\n"
            pynas_code += f"dim = [{width}, {height}]\n"
            pynas_code += "model.add_pbarl(pid=1, mid=1, Type={}, dim=dim, nsm=0.0)\n".format(
                '"BAR"'
            )
            pynas_code += "# pbarl.validate()\n\n\n"
        else:
            return
    elif mystran_writer.member.geos_shellthickness:
        # only use the first shellthickness object
        shellth_obj = mystran_writer.member.geos_shellthickness[0]["Object"]
        thickness = shellth_obj.Thickness.getValueAs("mm").Value
        pynas_code = "# pshell card, thin shell element properties\n"
        pynas_code += "model.add_pshell(pid=1, mid1=1, t={}, mid2=1, mid3=1)\n\n\n".format(
            thickness
        )
    else:
        pynas_code = "# psolid card, defines solid element\n"
        pynas_code += "model.add_psolid(pid=1, mid=1)\n\n\n"

    # write the pyNastran code
    f.write(pynas_code)

    # execute pyNastran code to add data to the model
    # print(model.get_bdf_stats())
    exec(pynas_code)
    # print(model.get_bdf_stats())

    return model


pynas_code = """
# pshell card, thin shell element properties
model.add_pshell(1, mid1=1, t=0.3, mid2=1, mid3=1)


"""


##  @}



# SPDX-License-Identifier: LGPL-2.1-or-later
# /****************************************************************************
#                                                                           *
#    Copyright (c) 2025 Weston Schmidt <weston_schmidt@alumni.purdue.edu>   *
#                                                                           *
#    This file is part of FreeCAD.                                          *
#                                                                           *
#    FreeCAD is free software: you can redistribute it and/or modify it     *
#    under the terms of the GNU Lesser General Public License as            *
#    published by the Free Software Foundation, either version 2.1 of the   *
#    License, or (at your option) any later version.                        *
#                                                                           *
#    FreeCAD is distributed in the hope that it will be useful, but         *
#    WITHOUT ANY WARRANTY; without even the implied warranty of             *
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
#    Lesser General Public License for more details.                        *
#                                                                           *
#    You should have received a copy of the GNU Lesser General Public       *
#    License along with FreeCAD. If not, see                                *
#    <https://www.gnu.org/licenses/>.                                       *
#                                                                           *
# ***************************************************************************/

"""Mock classes for FreeCAD GUI testing."""

import builtins
from .MockGui import (
    MockQIcon,
    MockQTreeWidgetItem,
    MockPySideUic,
    MockGetDocument,
    MockAddModule,
    MockDoCommandSkip,
)

# Set up all FreeCAD GUI mocks for testing
# Always create mocks for consistent testing
Gui = type("MockGui", (), {})()
QtCore = type("MockQtCore", (), {})()
QtGui = type("MockQtGui", (), {})()

# Patch QtGui with our mock classes
QtGui.QIcon = MockQIcon
QtGui.QTreeWidgetItem = MockQTreeWidgetItem

# Mock the PySideUic if it doesn't exist
if not hasattr(Gui, "PySideUic"):
    Gui.PySideUic = MockPySideUic

# Mock additional Gui methods that might be missing
if not hasattr(Gui, "getDocument"):
    Gui.getDocument = MockGetDocument

# Mock Selection module
if not hasattr(Gui, "Selection"):
    Gui.Selection = type(
        "MockSelection",
        (),
        {
            "clearSelection": lambda *args: None,
            "addSelection": lambda *args: None,
            "getSelection": lambda *args: [],
        },
    )()

# Mock addModule method
if not hasattr(Gui, "addModule"):
    Gui.addModule = MockAddModule

# Mock doCommandSkip method
if not hasattr(Gui, "doCommandSkip"):
    Gui.doCommandSkip = MockDoCommandSkip

# Make QtCore, QtGui and Gui available in the global namespace
builtins.QtCore = QtCore
builtins.QtGui = QtGui
builtins.Gui = Gui
builtins.QIcon = MockQIcon

__all__ = []



from Base.Metadata import export
from Base.Persistence import Persistence

@export(
    Include="Mod/Spreadsheet/App/PropertyColumnWidths.h",
    Namespace="Spreadsheet",
    Constructor=True,
)
class PropertyColumnWidths(Persistence):
    """
    Internal spreadsheet object

    Author: Eivind Kvedalen (eivind@kvedalen.name)
    License: LGPL-2.1-or-later
    """



from Base.Metadata import export, constmethod
from Base.PyObjectBase import PyObjectBase
from Part.App.TopoShapeEdge import TopoShapeEdge
from Part.App.TopoShape import TopoShape
from typing import overload

@export(
    PythonName="Part.ShapeFix.EdgeConnect",
    Include="ShapeFix_EdgeConnect.hxx",
    Constructor=True,
    Delete=True,
)
class ShapeFix_EdgeConnect(PyObjectBase):
    """
    Root class for fixing operations

    Author: Werner Mayer (wmayer@users.sourceforge.net)
    Licence: LGPL
    """

    @overload
    def add(self, edge1: TopoShapeEdge, edge2: TopoShapeEdge) -> None: ...
    @overload
    def add(self, shape: TopoShape) -> None: ...
    def add(self, *args, **kwargs) -> None:
        """
        add(edge, edge)
        Adds information on connectivity between start vertex
        of second edge and end vertex of first edge taking
        edges orientation into account

        add(shape)
        Adds connectivity information for the whole shape.
        """
        ...

    def build(self) -> None:
        """
        Builds shared vertices, updates their positions and tolerances
        """
        ...

    def clear(self) -> None:
        """
        Clears internal data structure
        """
        ...



from Base.Metadata import export, constmethod
from Part.Geom2d.Curve2d import Curve2d
from typing import overload

@export(
    PythonName="Part.Geom2d.Line2d",
    Twin="Geom2dLine",
    TwinPointer="Geom2dLine",
    Include="Mod/Part/App/Geometry2d.h",
    FatherInclude="Mod/Part/App/Geom2d/Curve2dPy.h",
    Constructor=True,
)
class Line2d(Curve2d):
    """
    Describes an infinite line in 2D space
    To create a line there are several ways:
    Part.Geom2d.Line2d()
        Creates a default line.

    Part.Geom2d.Line2d(Line)
        Creates a copy of the given line.

    Part.Geom2d.Line2d(Point,Dir)
        Creates a line that goes through two given points.

    Author: Werner Mayer (wmayer@users.sourceforge.net)
    Licence: LGPL
    """

    Location: object = ...
    """Returns the location of this line."""

    Direction: object = ...
    """Returns the direction of this line."""



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 FreeCAD Developers                                 *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides functions to create non-parametric arrayed copies."""
## @package array
# \ingroup draftfunctions
# \brief Provides functions to create non-parametric arrayed copies.

## \addtogroup draftfunctions
# @{
import FreeCAD as App
import draftutils.utils as utils
import draftfunctions.move as move
import draftfunctions.rotate as rotate


def array(objectslist, arg1, arg2, arg3, arg4=None, arg5=None, arg6=None):
    """
    This function creates an array of independent objects.
    Use make_array() to create a parametric array object.

    Creates an array of the given objects (that can be an object or a list
    of objects).

    In case of rectangular array, xnum of iterations in the x direction
    at xvector distance between iterations, and same for y and z directions
    with yvector and ynum and zvector and znum.

    In case of polar array, center is a vector, totalangle is the angle
    to cover (in degrees) and totalnum is the number of objects, including
    the original.

    Use
    ---
    array(objectslist, xvector, yvector, xnum, ynum) for rectangular array

    array(objectslist, xvector, yvector, zvector, xnum, ynum, znum) for rectangular array

    array(objectslist, center, totalangle, totalnum) for polar array
    """

    if arg6:
        rectArray2(objectslist, arg1, arg2, arg3, arg4, arg5, arg6)
    elif arg4:
        rectArray(objectslist, arg1, arg2, arg3, arg4)
    else:
        polarArray(objectslist, arg1, arg2, arg3)


def rectArray(objectslist, xvector, yvector, xnum, ynum):
    utils.type_check(
        [(xvector, App.Vector), (yvector, App.Vector), (xnum, int), (ynum, int)], "rectArray"
    )
    if not isinstance(objectslist, list):
        objectslist = [objectslist]
    for xcount in range(xnum):
        currentxvector = App.Vector(xvector).multiply(xcount)
        if not xcount == 0:
            move.move(objectslist, currentxvector, True)
        for ycount in range(ynum):
            currentxvector = App.Vector(currentxvector)
            currentyvector = currentxvector.add(App.Vector(yvector).multiply(ycount))
            if not ycount == 0:
                move.move(objectslist, currentyvector, True)


def rectArray2(objectslist, xvector, yvector, zvector, xnum, ynum, znum):
    utils.type_check(
        [
            (xvector, App.Vector),
            (yvector, App.Vector),
            (zvector, App.Vector),
            (xnum, int),
            (ynum, int),
            (znum, int),
        ],
        "rectArray2",
    )
    if not isinstance(objectslist, list):
        objectslist = [objectslist]
    for xcount in range(xnum):
        currentxvector = App.Vector(xvector).multiply(xcount)
        if not xcount == 0:
            move.move(objectslist, currentxvector, True)
        for ycount in range(ynum):
            currentxvector = App.Vector(currentxvector)
            currentyvector = currentxvector.add(App.Vector(yvector).multiply(ycount))
            if not ycount == 0:
                move.move(objectslist, currentyvector, True)
            for zcount in range(znum):
                currentzvector = currentyvector.add(App.Vector(zvector).multiply(zcount))
                if not zcount == 0:
                    move.move(objectslist, currentzvector, True)


def polarArray(objectslist, center, angle, num):
    utils.type_check([(center, App.Vector), (num, int)], "polarArray")
    if not isinstance(objectslist, list):
        objectslist = [objectslist]
    fraction = float(angle) / num
    for i in range(num):
        currangle = fraction + (i * fraction)
        rotate.rotate(objectslist, currangle, center, copy=True)


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2020 Carlo Pavan <carlopav@gmail.com>                   *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the initialization code for the workbench's status bar.

The status bar is activated by `InitGui.py` when the workbench is started,
and is populated by various widgets, buttons and menus.
"""
## @package init_draft_statusbar
# \ingroup draftutils
# \brief Provides the initialization code for the workbench's status bar.

## \addtogroup draftutils
# @{
import PySide.QtCore as QtCore
import PySide.QtGui as QtGui
import PySide.QtWidgets as QtWidgets

import FreeCAD as App
import FreeCADGui as Gui
from draftutils import params
from draftutils.init_tools import get_draft_snap_commands
from draftutils.translate import translate

# ----------------------------------------------------------------------------
# SCALE WIDGET FUNCTIONS
# ----------------------------------------------------------------------------

draft_scales_metrics = [
    "1:1000",
    "1:500",
    "1:250",
    "1:200",
    "1:100",
    "1:50",
    "1:25",
    "1:20",
    "1:10",
    "1:5",
    "1:2",
    "1:1",
    "2:1",
    "5:1",
    "10:1",
    "20:1",
    translate("draft", "Custom"),
]

draft_scales_arch_imperial = [
    "1/16in=1ft",
    "3/32in=1ft",
    "1/8in=1ft",
    "3/16in=1ft",
    "1/4in=1ft",
    "3/8in=1ft",
    "1/2in=1ft",
    "3/4in=1ft",
    "1in=1ft",
    "1.5in=1ft",
    "3in=1ft",
    translate("draft", "Custom"),
]

draft_scales_eng_imperial = [
    "1in=10ft",
    "1in=20ft",
    "1in=30ft",
    "1in=40ft",
    "1in=50ft",
    "1in=60ft",
    "1in=70ft",
    "1in=80ft",
    "1in=90ft",
    "1in=100ft",
    translate("draft", "Custom"),
]


def get_scales(unit_system=0):
    """
    returns the list of preset scales according to unit system.

    Parameters:
    unit_system =   0 : default from user preferences
                    1 : metrics
                    2 : imperial architectural
                    3 : imperial engineering
    """

    if unit_system == 0:
        scale_units_system = params.get_param("UserSchema", path="Units")
        if scale_units_system in [2, 3, 5]:
            return draft_scales_arch_imperial
        elif scale_units_system in [7]:
            return draft_scales_eng_imperial
        else:
            return draft_scales_metrics
    elif unit_system == 1:
        return draft_scales_metrics
    elif unit_system == 2:
        return draft_scales_arch_imperial
    elif unit_system == 3:
        return draft_scales_eng_imperial


def scale_to_label(scale):
    """
    transform a float number into a 1:X or X:1 scale and return it as label
    """
    if scale <= 0:
        return "1:1"
    f = round(scale, 2)
    if f == 1:
        return "1:1"
    if f > 1:
        f = f.as_integer_ratio()
        if f[1] == 1:
            return str(f[0]) + ":1"
        return str(scale)
    f = round(1 / scale, 2)
    f = f.as_integer_ratio()
    if f[1] == 1:
        return "1:" + str(f[0])
    return str(scale)


def label_to_scale(label):
    """
    transform a scale string into scale factor as float
    """
    try:
        scale = float(label)
        return scale
    except Exception:
        if ":" in label:
            f = label.split(":")
        elif "=" in label:
            f = label.split("=")
        else:
            return None
        if len(f) == 2:
            try:
                num = App.Units.Quantity(f[0]).Value
                den = App.Units.Quantity(f[1]).Value
                scale = num / den
                return scale
            except Exception:
                err = translate("draft", "Unable to convert input into a scale factor")
                App.Console.PrintWarning(err)
                return None


def _set_scale(action):
    """
    triggered by scale pushbutton, set DefaultAnnoScaleMultiplier in preferences
    """
    mw = Gui.getMainWindow()
    sb = mw.statusBar()

    scale_widget = sb.findChild(QtWidgets.QToolBar, "draft_scale_widget")

    if action.text() == translate("draft", "Custom"):
        title_text = translate("draft", "Set Custom Scale")
        dialog_text = translate("draft", "Set custom annotation scale in format x:x, x=x")
        custom_scale = QtWidgets.QInputDialog.getText(None, title_text, dialog_text)
        if custom_scale[1]:
            print(custom_scale[0])
            scale = label_to_scale(custom_scale[0])
            if scale is None:
                return
            if scale <= 0:
                return
            params.set_param("DefaultAnnoScaleMultiplier", 1 / scale)
            cs = scale_to_label(scale)
            scale_widget.scaleLabel.setText(cs)
    else:
        text_scale = action.text()
        scale_widget.scaleLabel.setText(text_scale)
        scale = label_to_scale(text_scale)
        params.set_param("DefaultAnnoScaleMultiplier", 1 / scale)


# ----------------------------------------------------------------------------
# MAIN DRAFT STATUSBAR FUNCTIONS
# ----------------------------------------------------------------------------


def init_draft_statusbar_scale():
    """
    this function initializes draft statusbar scale widget
    """
    mw = Gui.getMainWindow()
    sb = mw.statusBar()

    scale_widget = QtWidgets.QToolBar()
    # prevent the widget from showing up in the toolbar area context menu:
    scale_widget.toggleViewAction().setVisible(False)
    scale_widget.setObjectName("draft_scale_widget")
    text = translate(
        "draft",
        "Draft Scale Widget",
        "A context menu action used to show or hide this toolbar widget",
    )
    scale_widget.setWindowTitle(text)

    # get scales list according to system units
    draft_scales = get_scales()

    # get draft annotation scale
    scale_multiplier = params.get_param("DefaultAnnoScaleMultiplier")
    annotation_scale = 1 / scale_multiplier if scale_multiplier > 0 else 1

    # initializes scale widget
    scale_widget.draft_scales = draft_scales
    scaleLabel = QtWidgets.QPushButton("Scale")
    scaleLabel.setObjectName("ScaleLabel")
    scaleLabel.setFlat(True)
    menu = QtWidgets.QMenu(scaleLabel)
    gUnits = QtGui.QActionGroup(menu)
    for u in draft_scales:
        a = QtGui.QAction(gUnits)
        a.setText(u)
        menu.addAction(a)
    scaleLabel.setMenu(menu)
    gUnits.triggered.connect(_set_scale)
    scale_label = scale_to_label(annotation_scale)
    scaleLabel.setText(scale_label)
    scaleLabel.setToolTip(translate("draft", "Set the scale used by Draft annotation tools"))
    scale_widget.addWidget(scaleLabel)
    scale_widget.scaleLabel = scaleLabel

    # add scale widget to the statusbar
    sb.insertPermanentWidget(3, scale_widget)
    scale_widget.show()


def init_draft_statusbar_snap():
    """
    this function initializes draft statusbar snap widget
    """

    def _spacer():
        """
        empty label instead of snap_widget.addSeparator()
        """
        label = QtWidgets.QLabel()
        label.setFixedWidth(2)
        return label

    mw = Gui.getMainWindow()
    sb = mw.statusBar()

    # check if the toolbar is available, without it the required actions
    # may be missing:
    if mw.findChild(QtWidgets.QToolBar, "Draft Snap") is None:
        return

    # snap widget:
    snap_widget = QtWidgets.QToolBar()
    # prevent the widget from showing up in the toolbar area context menu:
    snap_widget.toggleViewAction().setVisible(False)
    snap_widget.setObjectName("draft_snap_widget")
    text = translate(
        "draft",
        "Draft Snap Widget",
        "A context menu action used to show or hide this toolbar widget",
    )
    snap_widget.setWindowTitle(text)
    snap_widget.setOrientation(QtCore.Qt.Orientation.Horizontal)
    snap_widget.setIconSize(QtCore.QSize(16, 16))
    sb.insertPermanentWidget(2, snap_widget)

    # grid button:
    snap_widget.addAction(Gui.Command.get("Draft_ToggleGrid").getAction()[0])

    snap_widget.addWidget(_spacer())

    # lock button:
    snap_widget.addAction(Gui.Command.get("Draft_Snap_Lock").getAction()[0])
    snap_action = snap_widget.children()[-1]
    snap_action.setFixedWidth(40)  # Widen the button.

    snap_widget.addWidget(_spacer())

    # dimension button:
    snap_widget.addAction(Gui.Command.get("Draft_Snap_Dimensions").getAction()[0])

    snap_widget.addWidget(_spacer())

    # ortho button:
    snap_widget.addAction(Gui.Command.get("Draft_Snap_Ortho").getAction()[0])

    snap_widget.addWidget(_spacer())

    # working plane button:
    snap_widget.addAction(Gui.Command.get("Draft_Snap_WorkingPlane").getAction()[0])

    # menu for lock button:
    for cmd in get_draft_snap_commands():
        if cmd not in [
            "Separator",
            "Draft_ToggleGrid",
            "Draft_Snap_Lock",  # Is automatically added to the menu anyway.
            "Draft_Snap_Dimensions",
            "Draft_Snap_Ortho",
            "Draft_Snap_WorkingPlane",
        ]:
            snap_action.addAction(Gui.Command.get(cmd).getAction()[0])


def show_draft_statusbar_scale():
    """
    shows draft statusbar scale widget
    """
    mw = Gui.getMainWindow()
    sb = mw.statusBar()

    scale_widget = sb.findChild(QtWidgets.QToolBar, "draft_scale_widget")
    if scale_widget:
        scale_widget.show()
    else:
        scale_widget = mw.findChild(QtWidgets.QToolBar, "draft_scale_widget")
        if scale_widget:
            sb.insertPermanentWidget(3, scale_widget)
            scale_widget.show()
        else:
            init_draft_statusbar_scale()


def show_draft_statusbar_snap():
    """
    shows draft statusbar snap widget
    """
    mw = Gui.getMainWindow()
    sb = mw.statusBar()

    snap_widget = sb.findChild(QtWidgets.QToolBar, "draft_snap_widget")
    if snap_widget:
        snap_widget.setOrientation(QtCore.Qt.Orientation.Horizontal)
        snap_widget.show()
    else:
        snap_widget = mw.findChild(QtWidgets.QToolBar, "draft_snap_widget")
        if snap_widget:
            sb.insertPermanentWidget(2, snap_widget)
            snap_widget.setOrientation(QtCore.Qt.Orientation.Horizontal)
            snap_widget.show()
        else:
            init_draft_statusbar_snap()


def hide_draft_statusbar_scale():
    """
    hides draft statusbar scale widget
    """
    mw = Gui.getMainWindow()
    sb = mw.statusBar()

    scale_widget = sb.findChild(QtWidgets.QToolBar, "draft_scale_widget")
    if scale_widget is None:
        # when switching workbenches, the toolbar sometimes "jumps"
        # out of the status bar to any other dock area...
        scale_widget = mw.findChild(QtWidgets.QToolBar, "draft_scale_widget")
    if scale_widget:
        scale_widget.hide()


def hide_draft_statusbar_snap():
    """
    hides draft statusbar snap widget
    """
    mw = Gui.getMainWindow()
    sb = mw.statusBar()

    snap_widget = sb.findChild(QtWidgets.QToolBar, "draft_snap_widget")
    if snap_widget is None:
        # when switching workbenches, the toolbar sometimes "jumps"
        # out of the status bar to any other dock area...
        snap_widget = mw.findChild(QtWidgets.QToolBar, "draft_snap_widget")
    if snap_widget:
        snap_widget.hide()


def show_draft_statusbar():
    """
    shows draft statusbar if present or initializes it
    """
    if params.get_param("DisplayStatusbarScaleWidget"):
        QtCore.QTimer().singleShot(500, show_draft_statusbar_scale)
    if params.get_param("DisplayStatusbarSnapWidget"):
        QtCore.QTimer().singleShot(500, show_draft_statusbar_snap)


def hide_draft_statusbar():
    """
    hides draft statusbar if present
    """
    # Delay required in case the Draft WB is autoloaded,
    # else show_draft_statusbar will not yet be done.
    QtCore.QTimer().singleShot(500, hide_draft_statusbar_scale)
    QtCore.QTimer().singleShot(500, hide_draft_statusbar_snap)


## @}
